/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */


/************************************************************************
 * file.h: declaration of the File class
 *
 * A File object represents a file (most likely a named pipe) bound
 * to a thread.
************************************************************************/
#ifndef _thread_src_file_h_
#define _thread_src_file_h_


struct File : public IOBindable
{
	PDDESC	fd;

	static File* Find( PDDESC fd, thread_t* tidp = NULL );
	static File* Alloc( PDDESC fd,
						thread_t tid,
						bool special,
						int (*will_block)(void*),
						void* desc,
						thread_t* pfmid );
	static bool	IsValid( PDDESC fd );

	void	Init( thread_t tid,
					PDDESC filedesc,
					bool special,
					int (*will_block)(void*),
					void* desc )
				{
					IOBindable::Init( tid, special, will_block, desc );
					fd = filedesc;
				}

	virtual	tag_t	GetTag( void ) const			{ return MSG_TAG_FILE; }
	virtual	PDDESC	GetWaiterDesc( void ) const		{ return fd; }
	virtual	int		Read( void* buf, unsigned bufsize, unsigned* nbytes );
	virtual	int		Write( void* buf, unsigned bufsize, unsigned* nbytes );

#if defined(THREADLIB_DEBUG)
	virtual	void	WriteTo( ostream& s ) const;
#endif // defined(THREADLIB_DEBUG)
};


#endif // defined(_thread_src_file_h)
